"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
var __spread = (this && this.__spread) || function () {
    for (var ar = [], i = 0; i < arguments.length; i++) ar = ar.concat(__read(arguments[i]));
    return ar;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var codemirror_1 = __importDefault(require("codemirror"));
var clonedeep_1 = __importDefault(require("clonedeep"));
function pushPatternContext(state, pattern) {
    state.patternContext = {
        pre: state.patternContext,
        pattern: pattern,
    };
}
function popPatternContext(state) {
    state.patternContext = state.patternContext.pre;
}
function pushStateContext(config, state, mode, stream) {
    var modeObj = codemirror_1.default.getMode(config, mode);
    var context = {
        pre: null,
        mode: modeObj,
        state: codemirror_1.default.startState(modeObj, stream && stream.indentation()),
    };
    if (state.stateContext) {
        context.pre = state.stateContext;
    }
    state.stateContext = context;
}
function popStateContext(state) {
    state.stateContext = state.stateContext.pre;
}
function getLocalMode(state) {
    return state.stateContext && state.stateContext.mode || null;
}
function getLocalState(state) {
    return state.stateContext && state.stateContext.state || null;
}
function getCurrentPattern(state) {
    var patternContext = state.patternContext;
    return patternContext.pattern;
}
var AnchorMode;
(function (AnchorMode) {
    AnchorMode[AnchorMode["NONE"] = 0] = "NONE";
    AnchorMode[AnchorMode["START"] = 1] = "START";
    AnchorMode[AnchorMode["END"] = 2] = "END";
    AnchorMode[AnchorMode["ALL"] = 3] = "ALL";
})(AnchorMode || (AnchorMode = {}));
function makeSureAnchor(source, start, exist) {
    if (start) {
        if (exist && !source.startsWith('^')) {
            return "^" + source;
        }
        else if (!exist && source.startsWith('^')) {
            return source.slice(1);
        }
    }
    else {
        if (exist && !source.endsWith('$')) {
            return source + "$";
        }
        else if (!exist && source.endsWith('$')) {
            return source.slice(0, -1);
        }
    }
    return source;
}
function getCachedRegExp(state, regExp, anchor, global) {
    var source = regExp.source, flags = regExp.flags;
    switch (anchor) {
        case AnchorMode.ALL:
            source = makeSureAnchor(source, true, true);
            source = makeSureAnchor(source, false, true);
            break;
        case AnchorMode.END:
            source = makeSureAnchor(source, true, false);
            source = makeSureAnchor(source, false, true);
            break;
        case AnchorMode.NONE:
            source = makeSureAnchor(source, true, false);
            source = makeSureAnchor(source, false, false);
            break;
        case AnchorMode.START:
            source = makeSureAnchor(source, true, true);
            source = makeSureAnchor(source, false, false);
    }
    if (global && flags.indexOf('g') === -1) {
        flags = flags + "g";
    }
    else if (!global && flags.indexOf('g') !== -1) {
        flags = flags.replace(/g/g, '');
    }
    var key = source + "tVes#aE$" + flags;
    var cached = state.regExpCache[key];
    if (!cached) {
        cached = new RegExp(source, flags);
        state.regExpCache[key] = cached;
    }
    return cached;
}
var MatchMode;
(function (MatchMode) {
    MatchMode[MatchMode["DEFAULT"] = 0] = "DEFAULT";
    MatchMode[MatchMode["PREFIX"] = 1] = "PREFIX";
})(MatchMode || (MatchMode = {}));
var REG_SPACE = /^\s*$/;
var REG_WORD = /\w/;
function match(line, position, mode, pattern, state) {
    if (typeof pattern === 'string') {
        if (mode === MatchMode.DEFAULT) {
            var index = line.indexOf(pattern, position);
            return [index, index >= 0 ? pattern : null];
        }
        else {
            return line.endsWith(pattern, position) ? [position - pattern.length, pattern] : [-1, null];
        }
    }
    if (typeof pattern === 'function') {
        return pattern(line, position, state);
    }
    if (mode === MatchMode.DEFAULT) {
        var reg = getCachedRegExp(state, pattern, AnchorMode.NONE, true);
        reg.lastIndex = position;
        var result = reg.exec(line);
        return result ? [result.index, result[0]] : [-1, null];
    }
    else {
        var reg = getCachedRegExp(state, pattern, AnchorMode.END, false);
        reg.lastIndex = 0;
        var result = reg.exec(line.substring(0, position));
        return result ? [result.index, result[0]] : [-1, null];
    }
}
function matchWithEscape(line, position, pattern, escape, state) {
    var _a;
    var _b = __read(match(line, position, MatchMode.DEFAULT, pattern, state), 2), pos = _b[0], matched = _b[1];
    var escaped = false;
    while (pos !== -1) {
        escaped = checkEscape(state, line, pos, escape);
        if (!escaped) {
            break;
        }
        var next = pos + (matched ? matched.length : 0);
        _a = __read(match(line, next, MatchMode.DEFAULT, pattern, state), 2), pos = _a[0], matched = _a[1];
    }
    pos = escaped ? -1 : pos;
    return [pos, pos !== -1 ? matched : null];
}
function checkEscape(state, line, offset, escape) {
    if (!escape)
        return false;
    var _a = __read(match(line, offset, MatchMode.PREFIX, escape, state), 1), pos = _a[0];
    return pos >= 0;
}
var FoundMode;
(function (FoundMode) {
    FoundMode[FoundMode["PUSH"] = 0] = "PUSH";
    FoundMode[FoundMode["POP"] = 1] = "POP";
    FoundMode[FoundMode["NONE"] = 2] = "NONE";
})(FoundMode || (FoundMode = {}));
function createLayers(parserConfig, state, line, offset) {
    var _a, e_1, _b, _c;
    var beforeEnter = parserConfig.beforeEnter, afterEnter = parserConfig.afterEnter, beforeExit = parserConfig.beforeExit, afterExit = parserConfig.afterExit;
    var layers = [];
    var matched = null;
    while (offset < line.length) {
        var found = FoundMode.NONE;
        var pos = -1;
        var tmpPos = void 0;
        var tmpMatched = null;
        var pattern = void 0;
        var prePattern = void 0;
        var nextPattern = void 0;
        prePattern = nextPattern = getCurrentPattern(state);
        var close_1 = prePattern.close, children = prePattern.children, escape_1 = prePattern.escape;
        if (close_1) {
            _a = __read(matchWithEscape(line, offset, close_1, escape_1, state), 2), tmpPos = _a[0], tmpMatched = _a[1];
            if (tmpPos >= 0) {
                found = FoundMode.POP;
                pos = tmpPos;
                matched = tmpMatched;
            }
        }
        if (children) {
            try {
                for (var children_1 = (e_1 = void 0, __values(children)), children_1_1 = children_1.next(); !children_1_1.done; children_1_1 = children_1.next()) {
                    var child = children_1_1.value;
                    var open_1 = child.open;
                    if (open_1) {
                        _c = __read(matchWithEscape(line, offset, open_1, escape_1, state), 2), tmpPos = _c[0], tmpMatched = _c[1];
                        if (tmpPos >= 0 && (pos === -1 || tmpPos < pos)) {
                            found = FoundMode.PUSH;
                            pos = tmpPos;
                            matched = tmpMatched;
                            nextPattern = child;
                        }
                    }
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (children_1_1 && !children_1_1.done && (_b = children_1.return)) _b.call(children_1);
                }
                finally { if (e_1) throw e_1.error; }
            }
        }
        if (found !== FoundMode.NONE) {
            state.textBefore += line.slice(offset, pos);
            if (found === FoundMode.PUSH) {
                pattern = nextPattern;
                beforeEnter && beforeEnter(state, { line: line, pos: pos, textBefore: state.textBefore, matched: matched, pattern: pattern });
                pushPatternContext(state, pattern);
                afterEnter && afterEnter(state, { line: line, pos: pos, textBefore: state.textBefore, matched: matched, pattern: pattern });
            }
            else {
                pattern = prePattern;
                beforeExit && beforeExit(state, { line: line, pos: pos, textBefore: state.textBefore, matched: matched, pattern: pattern });
                popPatternContext(state);
                afterExit && afterExit(state, { line: line, pos: pos, textBefore: state.textBefore, matched: matched, pattern: pattern });
                nextPattern = getCurrentPattern(state);
            }
            state.textBefore = '';
            layers.push({
                pos: pos,
                matched: matched,
                open: found === FoundMode.PUSH,
                prePattern: prePattern,
                nextPattern: nextPattern,
            });
            offset = pos + (matched ? matched.length : 0);
        }
        else {
            state.textBefore += line.slice(offset);
            break;
        }
    }
    return layers;
}
function syncState(config, state, stream, layer) {
    if (layer.open && layer.nextPattern.mode) {
        pushStateContext(config, state, layer.nextPattern.mode, stream);
    }
    else if (!layer.open && layer.prePattern.mode) {
        popStateContext(state);
    }
}
function tokenPattern(config, stream, state) {
    var layers = state.layers;
    var layer = layers && layers[0];
    if (!layer) {
        throw new Error('This is impossible!');
    }
    var open = layer.open, pos = layer.pos, matched = layer.matched, prePattern = layer.prePattern, nextPattern = layer.nextPattern;
    var pattern = open ? nextPattern : prePattern;
    var patternStyles = pattern.patternStyles;
    if (stream.pos !== pos || !patternStyles) {
        throw new Error('This is impossible!');
    }
    state.useRoot = true;
    stream.pos = pos + (matched ? matched.length : 0);
    syncState(config, state, stream, layer);
    layers.shift();
    if (stream.eol()) {
        state.textBefore += '\n';
    }
    return patternStyles[open ? 0 : 1];
}
function tokenUntil(config, stream, state, until, shift, syncStateWhenReach) {
    var localMode = getLocalMode(state);
    var localState = getLocalState(state);
    var layers = state.layers;
    var style;
    if (localMode && localMode.token) {
        style = localMode.token(stream, localState);
        if (stream.pos > until) {
            stream.backUp(stream.pos - until);
        }
    }
    else {
        style = null;
        stream.pos = until;
    }
    if (layers && layers[0]) {
        var layer = layers[0];
        if (stream.pos === until) {
            if (syncStateWhenReach) {
                syncState(config, state, stream, layer);
            }
            if (shift) {
                layers.shift();
            }
        }
    }
    if (stream.eol()) {
        state.textBefore += '\n';
    }
    return style;
}
function token(config, parserConfig, stream, state) {
    if (!state.start || stream.sol()) {
        state.start = true;
        state.useRoot = false;
        state.layers = createLayers(parserConfig, state, stream.string, stream.start);
    }
    var layer = state.layers && state.layers[0];
    if (layer) {
        var prePattern = layer.prePattern, nextPattern = layer.nextPattern, pos = layer.pos, matched = layer.matched, open_2 = layer.open;
        var end = pos + (matched ? matched.length : 0);
        if (stream.start < pos) {
            return tokenUntil(config, stream, state, pos, false, false);
        }
        else if (stream.start === pos) {
            var pattern = open_2 ? nextPattern : prePattern;
            var patternStyles = pattern.patternStyles, includePattern = pattern.includePattern;
            if (patternStyles) {
                return tokenPattern(config, stream, state);
            }
            else if (open_2 && includePattern) {
                nextPattern.mode && pushStateContext(config, state, nextPattern.mode, stream);
                return tokenUntil(config, stream, state, end, true, false);
            }
            else if (!open_2 && !includePattern) {
                prePattern.mode && popStateContext(state);
                return tokenUntil(config, stream, state, end, true, false);
            }
            else {
                return tokenUntil(config, stream, state, end, true, true);
            }
        }
        else if (stream.start < end) {
            return tokenUntil(config, stream, state, end, true, true);
        }
        else {
            throw new Error('This is impossible!');
        }
    }
    else {
        var localMode = getLocalMode(state);
        var localState = getLocalState(state);
        if (localMode && localMode.token) {
            return localMode.token(stream, localState);
        }
        else {
            stream.pos = stream.string.length;
            return null;
        }
    }
}
function copyPatternContext(context) {
    if (!context)
        return null;
    return {
        pre: copyPatternContext(context.pre),
        pattern: context.pattern,
    };
}
function copyStateContext(context) {
    if (!context)
        return null;
    return {
        pre: copyStateContext(context.pre),
        mode: context.mode,
        state: codemirror_1.default.copyState(context.mode, context.state),
    };
}
var FLAG_KEYWORDS = ['if'];
codemirror_1.default.defineMode('cxj-template-flag', function () {
    return {
        token: function (stream) {
            if (stream.eatSpace()) {
                return null;
            }
            if (stream.eatWhile(REG_WORD)) {
                var matched = stream.current();
                var idx = FLAG_KEYWORDS.indexOf(matched);
                return idx >= 0 ? "keyword cxj-template-flag-keyword cxj-template-flag-keyword-" + matched : null;
            }
            stream.next();
            return null;
        }
    };
});
codemirror_1.default.defineMode('cxj-template', function (config, parserConfig) {
    var mode = parserConfig.mode, codeMode = parserConfig.codeMode, open = parserConfig.open, close = parserConfig.close, customBeforeEnter = parserConfig.beforeEnter, customAfterExit = parserConfig.afterExit, rest = __rest(parserConfig, ["mode", "codeMode", "open", "close", "beforeEnter", "afterExit"]);
    var defaultOptions = createDefaultOptions(mode, codeMode);
    var defaultBeforeEnter = defaultOptions.beforeEnter, defaultAfterExit = defaultOptions.afterExit, restDefaultOptions = __rest(defaultOptions, ["beforeEnter", "afterExit"]);
    var finalParserConfig = __assign(__assign(__assign({}, restDefaultOptions), rest), { beforeEnter: function (state, info) {
            defaultBeforeEnter(state, info);
            customBeforeEnter && customBeforeEnter(state, info);
        },
        afterExit: function (state, info) {
            defaultAfterExit(state, info);
            customAfterExit && customAfterExit(state, info);
        } });
    var baseMode = finalParserConfig.mode;
    var baseModeObj = baseMode ? codemirror_1.default.getMode(config, baseMode) : null;
    var initialStateContext = baseModeObj ? {
        pre: null,
        mode: baseModeObj,
        state: codemirror_1.default.startState(baseModeObj),
    } : null;
    // noinspection JSUnusedGlobalSymbols
    var modeObj = {
        token: function (ss, state) { return token(config, finalParserConfig, ss, state); },
        startState: function () {
            return {
                textBefore: '',
                patternContext: {
                    pre: null,
                    pattern: finalParserConfig,
                },
                stateContext: initialStateContext,
                useRoot: false,
                start: false,
                regExpCache: {},
                customs: {},
            };
        },
        copyState: function (state) {
            return __assign(__assign({}, state), { layers: __spread(state.layers || []), patternContext: copyPatternContext(state.patternContext), stateContext: copyStateContext(state.stateContext), customs: clonedeep_1.default(state.customs) });
        },
        indent: function (state, textAfter, line) {
            var localMode = getLocalMode(state);
            if (localMode && localMode.indent) {
                var localState = getLocalState(state);
                return localMode.indent(localState, textAfter, line);
            }
            else {
                return codemirror_1.default.Pass;
            }
        },
        innerMode: function (state) {
            return {
                state: getLocalState(state) || state,
                mode: getLocalMode(state) || modeObj,
            };
        }
    };
    return modeObj;
});
function createDefaultOptions(baseMode, codeMode) {
    var singleQuote = {
        mode: null,
        open: '\'',
        close: '\'',
        escape: '\\',
    };
    var doubleQuote = {
        mode: null,
        open: '"',
        close: '"',
        escape: '\\',
    };
    var parenthesis = {
        mode: null,
        open: '(',
        close: ')',
    };
    var bracket = {
        mode: null,
        open: '[',
        close: ']',
    };
    var brace = {
        mode: null,
        open: '{',
        close: '}',
    };
    var children = [singleQuote, doubleQuote, parenthesis, bracket, brace];
    parenthesis.children = bracket.children = brace.children = children;
    var code = {
        mode: codeMode,
        open: /#{\s*:?/,
        close: '}',
        children: children,
        patternStyles: ['bracket cxj-code open', 'bracket cxj-code close'],
    };
    var templateHeader = {
        mode: 'cxj-template-flag',
        open: '#[',
        close: ']',
        patternStyles: ['bracket cxj-template cxj-template-flag open', 'bracket cxj-template cxj-template-flag close'],
    };
    var template = {
        mode: baseMode,
        open: function (text, from, state) {
            var pos = text.indexOf('[', from);
            if (state.customs.justExitTemplate !== true || (pos >= 0 && !REG_SPACE.test(state.textBefore + text.slice(from, pos)))) {
                pos = -1;
            }
            return pos >= 0 ? [pos, '['] : [-1, null];
        },
        close: ']',
        children: __spread([
            code
        ], children),
        patternStyles: ['bracket cxj-template open', 'bracket cxj-template close'],
    };
    return {
        name: 'cxj-template',
        mode: baseMode,
        codeMode: codeMode,
        afterExit: function (state, _a) {
            var pattern = _a.pattern;
            if (pattern === template || pattern === templateHeader) {
                state.customs.justExitTemplate = true;
            }
        },
        beforeEnter: function (state) {
            state.customs.justExitTemplate = false;
        },
        children: [
            singleQuote,
            doubleQuote,
            code,
            templateHeader,
            template,
        ],
    };
}
exports.createDefaultOptions = createDefaultOptions;
//# sourceMappingURL=index.js.map